// -- a single match of length 12 at chosen positions in a 19-mer,
//  allowing one transition 
// -- a single match of length 11 at chosen positions in a 18-mer,
//  allowing one transition 

#include "util.h"
#include "seq.h"
#include "dna.h"
#include "bz_table.h"
#include "bz_hit19.h"

#include "astack.c"
// #include "cache.c"

//              1.............56789
/* template     1110100110010101111 (12 of 19) */
/* partitioned: 11101001100101 1111
*               111			shifted 11 bp = 22 bits
*		    1			shifted 9 bp  = 18 bits
*		       11		shifted 5 bp  = 10 bits
*			   1		shifted 2 bp  = 4 bits
*			     1
*/

//              1.............5678
//              ..............
// template     111010010100110111  (11 of 18)
//              0x3f << (11*2)
//		    3 << (9*2)
//		       3 << (6*2)
//		       3 << (4*2)
//		       3 << (1*2)

enum {
	mask1 = (0x3F) << 22,
	mask2 = 3 << 18,
	mask3 = (0xF) << 10,
	mask4 = 3 << 4,
	mask5 = 3,

	m18a1 = 0x3f << (11*2),
	m18a2 = 3 << (9*2),
	m18a3 = 3 << (6*2),
	m18a4 = 3 << (4*2),
	m18a5 = 3 << (1*2),

	maski = (1 << 26) - 1, /* picks out the 13 right-most bp */
	maskj = (1 << 6) - 1,  /* picks out the 3 right-most bp */
};

/*
*  To construct the 12 bp string from the 2 sides, i and j, of the partition:
*	shift the first 3 bp (first string of 1's in i) right 2 bp (4 bits),
*	shift the next 1 bp (second string of 1's in i) right 1 bp (2 bits),
*	shift the next 2 bp (third string of 1's in i) left 1 bp (2 bits),
*	shift the next 1 bp (forth string of 1's in i) left 3 bp (6 bits),
*	shift the next 1 bp (fifth string of 1's in i) left 4 bp (8 bits),
*	take the 4 bp (8 bits) of j.
*/

static unsigned int map_19to12(unsigned int i, unsigned int j)
{
	return
		((i & mask1) >> 4) |
		((i & mask2) >> 2) |
		((i & mask3) << 2) |
		((i & mask4) << 6) |
		((i & mask5) << 8) |
		(j & maskj);
}

#define N4E12 16777216	/* 4^12 */

/* -------------   build table of W-tuples in the first sequence  ----------- */

typedef struct hash_ {
	int pos;		/* position where word starts in sequence 1 */
	struct hash_ *next;
} Hash;

struct blast_table {
	Hash **hashtab; /* pointers to positions in seq 1, indexed by 12-mer */
	Hash *pool;     /* pool of free nodes */
	int npool;      /* number of nodes available */
	int n_node;     /* index of next free node in pool */
};

/* add_word - add to the table a position in sequence 1 of a given word */
static void add_word(blast_table_t *bt, int ecode, int pos)
{
	Hash *h;

	if (bt->n_node >= bt->npool)
		fatal("blast hashtab overflowed (can't happen)");
	h = &(bt->pool[bt->n_node++]);
	h->pos = pos;
	h->next = bt->hashtab[ecode];
	bt->hashtab[ecode] = h;
}

blast_table_t *blast_1219_new(SEQ *seq, int W)
{
	/* In the 19-mer ending at the current position, ei is the ecode
	*  of the 14-mer on the left, and ej is ecode of the 4 bp on the
	*  right.  (The bp between them isn't used, i.e., has a 0 in the
	*  template.)
	*/
	int i;
	unsigned int ei, ej, ecode, tecode;
	const unsigned char *s;
	const int len1 = SEQ_LEN(seq);
	const unsigned char *const seq1 = SEQ_CHARS(seq);
	const unsigned char *const seq_end = seq1 + len1;
	blast_table_t *bt;

	if (W != 12)
		fatal("Cannot handle W != 12.");

	bt = ckalloc(sizeof(*bt));
	bt->hashtab = (Hash **) ckallocz(N4E12*sizeof(Hash *));
	bt->npool = (W+1)*(len1 - W + 1);
	bt->pool = (Hash *) ckalloc(bt->npool*sizeof(Hash));
	bt->n_node = 0;
	
	for (s = seq1; s < seq_end;  ) {
	restart:
		/* init */
		ei = 0;
		for (i = 1; i < 14 && s < seq_end; ++i) {
			const int e = fasta_encoding[*s++];
			if (e < 0) goto restart; /* hit a run of 'X's */
			ei = (ei << 2) | e;
		}
		if (++s >= seq_end) continue;
		ej = 0;
		for (i = 1; i < 4 && s < seq_end; ++i) {
			const int e = fasta_encoding[*s++];
			if (e < 0) goto restart; /* hit a run of 'X's */
			ej = (ej << 2) | e;
		}
		/* compute */
		for ( ; s < seq_end; ++s) {
			const int e = fasta_encoding[*s];
			if (e < 0) goto restart; /* hit a run of 'X's */
			ej = (ej << 2) | e;
			ei = (ei << 2) | ((ej >> (2*5)) & 0x3);
			ecode = map_19to12(ei, ej);
			add_word(bt, ecode, s-seq1+1);
			for (i = 0; i < W; ++i) {
				// more general version:
				//j = (ecode & (3<<(2*i))) >> (2*i);
				//j = transition[j];
				//tecode = (ecode &~ (3<<(2*i))) | (j << (2*i));

				tecode = ecode ^ (2 << (2 * i)); /* A <--> G, C <--> T */
				add_word(bt, tecode, s-seq1+1);
			}
		} 
	}
	return bt;
}

void blast_1219_free(blast_table_t *bt)
{
	free(bt->hashtab);
	free(bt->pool);
}

/* -----------------------   search the second sequence   --------------------*/


msp_table_t *blast_1219_search(SEQ *seq1, SEQ *seq2, blast_table_t *bt, msp_table_t *mt, ss_t ss, int X, int K, int P)
{
	int i;
	Hash *h;
	const unsigned char *s;
	const unsigned char *const seq = SEQ_CHARS(seq2);
	const unsigned char *const seq_end = seq + SEQ_LEN(seq2);
	const int len1 = SEQ_LEN(seq1);
	const int len2 = SEQ_LEN(seq2);
	unsigned int ei, ej, ecode;
	static comb_t *diag_lev = 0;

        { unsigned int n = (len1+len2+1)*sizeof(int);
          n = roundup(n, 64*1024);
          diag_lev = comb_resize(diag_lev,n,len1);
        }

	for (s = seq; s < seq_end;  ) {
	restart:
		/* init */
		ei = 0;
		for (i = 1; i < 14 && s < seq_end; ++i) {
			const int e = fasta_encoding[*s++];
			if (e < 0) goto restart; /* hit a run of 'X's */
			ei = (ei << 2) | e;
		}
		if (++s >= seq_end) continue;
		ej = 0;
		for (i = 1; i < 4 && s < seq_end; ++i) {
			const int e = fasta_encoding[*s++];
			if (e < 0) goto restart; /* hit a run of 'X's */
			ej = (ej << 2) | e;
		}
		/* compute */
		for ( ; s < seq_end; ++s) {
			const int e = fasta_encoding[*s];
			if (e < 0) goto restart; /* hit a run of 'X's */
			ej = (ej << 2) | e;
			ei = (ei << 2) | ((ej >> (2*5)) & 0x3);
			ecode = map_19to12(ei, ej);
			for (h = bt->hashtab[ecode]; h; h = h->next) {
				msp_extend_hit(mt, seq1, seq2,
                                                ss, X, K, 19, P,
                                                h->pos, s-seq+1, diag_lev);
			}
		} 
	}
	comb_clear(diag_lev);
	return mt;
}
